@@ -4,12 +4,14 @@ import json
 from collections.abc import Sequence
 from dataclasses import dataclass
 from datetime import timezone
-from typing import Any, Union, cast
+from typing import Any, cast
 
 import pytest
 from inline_snapshot import snapshot
 
 from pydantic_ai import Agent, ModelHTTPError, ModelRetry
+from pydantic_ai.builtin_tools import WebSearchTool
+from pydantic_ai.exceptions import UserError
 from pydantic_ai.messages import (
     ImageUrl,
     ModelRequest,
@@ -17,14 +19,15 @@ from pydantic_ai.messages import (
     RetryPromptPart,
     SystemPromptPart,
     TextPart,
+    ThinkingPart,
     ToolCallPart,
     ToolReturnPart,
     UserPromptPart,
 )
 from pydantic_ai.tools import RunContext
-from pydantic_ai.usage import Usage
+from pydantic_ai.usage import RequestUsage, RunUsage
 
-from ..conftest import IsDatetime, IsNow, raise_if_exception, try_import
+from ..conftest import IsDatetime, IsInstance, IsNow, raise_if_exception, try_import
 
 with try_import() as imports_successful:
     import cohere
@@ -41,8 +44,7 @@ with try_import() as imports_successful:
     from pydantic_ai.models.cohere import CohereModel
     from pydantic_ai.providers.cohere import CohereProvider
 
-    # note: we use Union here for compatibility with Python 3.9
-    MockChatResponse = Union[ChatResponse, Exception]
+    MockChatResponse = ChatResponse | Exception
 
 pytestmark = [
     pytest.mark.skipif(not imports_successful(), reason='cohere not installed'),
@@ -101,29 +103,29 @@ async def test_request_simple_success(allow_model_requests: None):
 
     result = await agent.run('hello')
     assert result.output == 'world'
-    assert result.usage() == snapshot(Usage(requests=1))
+    assert result.usage() == snapshot(RunUsage(requests=1))
 
     # reset the index so we get the same response again
     mock_client.index = 0  # type: ignore
 
     result = await agent.run('hello', message_history=result.new_messages())
     assert result.output == 'world'
-    assert result.usage() == snapshot(Usage(requests=1))
+    assert result.usage() == snapshot(RunUsage(requests=1))
     assert result.all_messages() == snapshot(
         [
             ModelRequest(parts=[UserPromptPart(content='hello', timestamp=IsNow(tz=timezone.utc))]),
             ModelResponse(
                 parts=[TextPart(content='world')],
-                usage=Usage(requests=1),
                 model_name='command-r7b-12-2024',
                 timestamp=IsNow(tz=timezone.utc),
+                provider_name='cohere',
             ),
             ModelRequest(parts=[UserPromptPart(content='hello', timestamp=IsNow(tz=timezone.utc))]),
             ModelResponse(
                 parts=[TextPart(content='world')],
-                usage=Usage(requests=1),
                 model_name='command-r7b-12-2024',
                 timestamp=IsNow(tz=timezone.utc),
+                provider_name='cohere',
             ),
         ]
     )
@@ -147,11 +149,10 @@ async def test_request_simple_usage(allow_model_requests: None):
     result = await agent.run('Hello')
     assert result.output == 'world'
     assert result.usage() == snapshot(
-        Usage(
+        RunUsage(
             requests=1,
-            request_tokens=1,
-            response_tokens=1,
-            total_tokens=2,
+            input_tokens=1,
+            output_tokens=1,
             details={
                 'input_tokens': 1,
                 'output_tokens': 1,
@@ -191,9 +192,9 @@ async def test_request_structured_response(allow_model_requests: None):
                         tool_call_id='123',
                     )
                 ],
-                usage=Usage(requests=1),
                 model_name='command-r7b-12-2024',
                 timestamp=IsNow(tz=timezone.utc),
+                provider_name='cohere',
             ),
             ModelRequest(
                 parts=[
@@ -278,9 +279,9 @@ async def test_request_tool_call(allow_model_requests: None):
                         tool_call_id='1',
                     )
                 ],
-                usage=Usage(requests=1, total_tokens=0, details={}),
                 model_name='command-r7b-12-2024',
                 timestamp=IsNow(tz=timezone.utc),
+                provider_name='cohere',
             ),
             ModelRequest(
                 parts=[
@@ -300,15 +301,10 @@ async def test_request_tool_call(allow_model_requests: None):
                         tool_call_id='2',
                     )
                 ],
-                usage=Usage(
-                    requests=1,
-                    request_tokens=5,
-                    response_tokens=3,
-                    total_tokens=8,
-                    details={'input_tokens': 4, 'output_tokens': 2},
-                ),
+                usage=RequestUsage(input_tokens=5, output_tokens=3, details={'input_tokens': 4, 'output_tokens': 2}),
                 model_name='command-r7b-12-2024',
                 timestamp=IsNow(tz=timezone.utc),
+                provider_name='cohere',
             ),
             ModelRequest(
                 parts=[
@@ -322,19 +318,19 @@ async def test_request_tool_call(allow_model_requests: None):
             ),
             ModelResponse(
                 parts=[TextPart(content='final response')],
-                usage=Usage(requests=1),
                 model_name='command-r7b-12-2024',
                 timestamp=IsNow(tz=timezone.utc),
+                provider_name='cohere',
             ),
         ]
     )
     assert result.usage() == snapshot(
-        Usage(
+        RunUsage(
             requests=3,
-            request_tokens=5,
-            response_tokens=3,
-            total_tokens=8,
+            input_tokens=5,
+            output_tokens=3,
             details={'input_tokens': 4, 'output_tokens': 2},
+            tool_calls=1,
         )
     )
 
@@ -400,15 +396,103 @@ async def test_cohere_model_instructions(allow_model_requests: None, co_api_key:
                         content="The capital of France is Paris. It is the country's largest city and serves as the economic, cultural, and political center of France. Paris is known for its rich history, iconic landmarks such as the Eiffel Tower and the Louvre Museum, and its significant influence on fashion, cuisine, and the arts."
                     )
                 ],
-                usage=Usage(
-                    requests=1,
-                    request_tokens=542,
-                    response_tokens=63,
-                    total_tokens=605,
-                    details={'input_tokens': 13, 'output_tokens': 61},
+                usage=RequestUsage(
+                    input_tokens=542, output_tokens=63, details={'input_tokens': 13, 'output_tokens': 61}
                 ),
                 model_name='command-r7b-12-2024',
                 timestamp=IsDatetime(),
+                provider_name='cohere',
             ),
         ]
     )
+
+
+@pytest.mark.vcr()
+async def test_cohere_model_thinking_part(allow_model_requests: None, co_api_key: str, openai_api_key: str):
+    with try_import() as imports_successful:
+        from pydantic_ai.models.openai import OpenAIResponsesModel, OpenAIResponsesModelSettings
+        from pydantic_ai.providers.openai import OpenAIProvider
+
+    if not imports_successful():  # pragma: no cover
+        pytest.skip('OpenAI is not installed')
+
+    openai_model = OpenAIResponsesModel('o3-mini', provider=OpenAIProvider(api_key=openai_api_key))
+    co_model = CohereModel('command-r7b-12-2024', provider=CohereProvider(api_key=co_api_key))
+    agent = Agent(openai_model)
+
+    # We call OpenAI to get the thinking parts, because Google disabled the thoughts in the API.
+    # See https://github.com/pydantic/pydantic-ai/issues/793 for more details.
+    result = await agent.run(
+        'How do I cross the street?',
+        model_settings=OpenAIResponsesModelSettings(
+            openai_reasoning_effort='high', openai_reasoning_summary='detailed'
+        ),
+    )
+    assert result.all_messages() == snapshot(
+        [
+            ModelRequest(parts=[UserPromptPart(content='How do I cross the street?', timestamp=IsDatetime())]),
+            ModelResponse(
+                parts=[
+                    IsInstance(ThinkingPart),
+                    IsInstance(ThinkingPart),
+                    IsInstance(ThinkingPart),
+                    IsInstance(ThinkingPart),
+                    IsInstance(TextPart),
+                ],
+                usage=RequestUsage(input_tokens=13, output_tokens=1909, details={'reasoning_tokens': 1472}),
+                model_name='o3-mini-2025-01-31',
+                timestamp=IsDatetime(),
+                provider_name='openai',
+                provider_response_id='resp_680739f4ad748191bd11096967c37c8b048efc3f8b2a068e',
+            ),
+        ]
+    )
+
+    result = await agent.run(
+        'Considering the way to cross the street, analogously, how do I cross the river?',
+        model=co_model,
+        message_history=result.all_messages(),
+    )
+    assert result.all_messages() == snapshot(
+        [
+            ModelRequest(parts=[UserPromptPart(content='How do I cross the street?', timestamp=IsDatetime())]),
+            ModelResponse(
+                parts=[
+                    IsInstance(ThinkingPart),
+                    IsInstance(ThinkingPart),
+                    IsInstance(ThinkingPart),
+                    IsInstance(ThinkingPart),
+                    IsInstance(TextPart),
+                ],
+                usage=RequestUsage(input_tokens=13, output_tokens=1909, details={'reasoning_tokens': 1472}),
+                model_name='o3-mini-2025-01-31',
+                timestamp=IsDatetime(),
+                provider_name='openai',
+                provider_response_id='resp_680739f4ad748191bd11096967c37c8b048efc3f8b2a068e',
+            ),
+            ModelRequest(
+                parts=[
+                    UserPromptPart(
+                        content='Considering the way to cross the street, analogously, how do I cross the river?',
+                        timestamp=IsDatetime(),
+                    )
+                ]
+            ),
+            ModelResponse(
+                parts=[IsInstance(TextPart)],
+                usage=RequestUsage(
+                    input_tokens=1457, output_tokens=807, details={'input_tokens': 954, 'output_tokens': 805}
+                ),
+                model_name='command-r7b-12-2024',
+                timestamp=IsDatetime(),
+                provider_name='cohere',
+            ),
+        ]
+    )
+
+
+async def test_cohere_model_builtin_tools(allow_model_requests: None, co_api_key: str):
+    m = CohereModel('command-r7b-12-2024', provider=CohereProvider(api_key=co_api_key))
+    agent = Agent(m, builtin_tools=[WebSearchTool()])
+    with pytest.raises(UserError, match='Cohere does not support built-in tools'):
+        await agent.run('Hello')
