@@ -2,11 +2,15 @@
 
 from __future__ import annotations as _annotations
 
+import asyncio
+import dataclasses
+import re
 from datetime import timezone
 from typing import Annotated, Any, Literal
 
 import pytest
 from annotated_types import Ge, Gt, Le, Lt, MaxLen, MinLen
+from anyio import Event
 from inline_snapshot import snapshot
 from pydantic import BaseModel, Field
 
@@ -26,7 +30,7 @@ from pydantic_ai.messages import (
     VideoUrl,
 )
 from pydantic_ai.models.test import TestModel, _chars, _JsonSchemaTestData  # pyright: ignore[reportPrivateUsage]
-from pydantic_ai.usage import Usage
+from pydantic_ai.usage import RequestUsage, RunUsage
 
 from ..conftest import IsNow, IsStr
 
@@ -102,7 +106,7 @@ def test_tool_retry():
             ModelRequest(parts=[UserPromptPart(content='Hello', timestamp=IsNow(tz=timezone.utc))]),
             ModelResponse(
                 parts=[ToolCallPart(tool_name='my_ret', args={'x': 0}, tool_call_id=IsStr())],
-                usage=Usage(requests=1, request_tokens=51, response_tokens=4, total_tokens=55),
+                usage=RequestUsage(input_tokens=51, output_tokens=4),
                 model_name='test',
                 timestamp=IsNow(tz=timezone.utc),
             ),
@@ -118,7 +122,7 @@ def test_tool_retry():
             ),
             ModelResponse(
                 parts=[ToolCallPart(tool_name='my_ret', args={'x': 0}, tool_call_id=IsStr())],
-                usage=Usage(requests=1, request_tokens=61, response_tokens=8, total_tokens=69),
+                usage=RequestUsage(input_tokens=61, output_tokens=8),
                 model_name='test',
                 timestamp=IsNow(tz=timezone.utc),
             ),
@@ -131,7 +135,7 @@ def test_tool_retry():
             ),
             ModelResponse(
                 parts=[TextPart(content='{"my_ret":"1"}')],
-                usage=Usage(requests=1, request_tokens=62, response_tokens=12, total_tokens=74),
+                usage=RequestUsage(input_tokens=62, output_tokens=12),
                 model_name='test',
                 timestamp=IsNow(tz=timezone.utc),
             ),
@@ -154,20 +158,50 @@ def test_output_tool_retry_error_handled():
         call_count += 1
         raise ModelRetry('Fail')
 
-    with pytest.raises(UnexpectedModelBehavior, match='Exceeded maximum retries'):
+    with pytest.raises(UnexpectedModelBehavior, match=re.escape('Exceeded maximum retries (2) for output validation')):
         agent.run_sync('Hello', model=TestModel())
 
     assert call_count == 3
 
 
-def test_output_tool_retry_error_handled_with_custom_args(set_event_loop: None):
+@dataclasses.dataclass
+class AgentRunDeps:
+    run_id: int
+
+
+@pytest.mark.anyio
+async def test_multiple_concurrent_tool_retries():
+    class OutputModel(BaseModel):
+        x: int
+        y: str
+
+    agent = Agent('test', deps_type=AgentRunDeps, output_type=OutputModel, retries=2)
+    retried_run_ids = set[int]()
+    event = Event()
+
+    run_ids = list(range(5))  # fire off 5 run ids that will all retry the tool before they finish
+
+    @agent.tool
+    async def tool_that_must_be_retried(ctx: RunContext[AgentRunDeps]) -> None:
+        if ctx.deps.run_id not in retried_run_ids:
+            retried_run_ids.add(ctx.deps.run_id)
+            raise ModelRetry('Fail')
+        if len(retried_run_ids) == len(run_ids):  # pragma: no branch  # won't branch if all runs happen very quickly
+            event.set()
+        await event.wait()  # ensure a retry is done by all runs before any of them finish their flow
+        return None
+
+    await asyncio.gather(*[agent.run('Hello', model=TestModel(), deps=AgentRunDeps(run_id)) for run_id in run_ids])
+
+
+def test_output_tool_retry_error_handled_with_custom_args():
     class ResultModel(BaseModel):
         x: int
         y: str
 
     agent = Agent('test', output_type=ResultModel, retries=2)
 
-    with pytest.raises(UnexpectedModelBehavior, match='Exceeded maximum retries'):
+    with pytest.raises(UnexpectedModelBehavior, match=r'Exceeded maximum retries \(2\) for output validation'):
         agent.run_sync('Hello', model=TestModel(custom_output_args={'foo': 'a', 'bar': 1}))
 
 
@@ -303,6 +337,6 @@ def test_max_items():
 )
 def test_different_content_input(content: AudioUrl | VideoUrl | ImageUrl | BinaryContent):
     agent = Agent()
-    result = agent.run_sync('x', model=TestModel(custom_output_text='custom'))
+    result = agent.run_sync(['x', content], model=TestModel(custom_output_text='custom'))
     assert result.output == snapshot('custom')
-    assert result.usage() == snapshot(Usage(requests=1, request_tokens=51, response_tokens=1, total_tokens=52))
+    assert result.usage() == snapshot(RunUsage(requests=1, input_tokens=51, output_tokens=1))
