@@ -3,10 +3,10 @@
 WARNING: running these tests will make use of the relevant API tokens (and cost money).
 """
 
+import json
 import os
-from collections.abc import AsyncIterator
+from collections.abc import AsyncIterator, Callable
 from pathlib import Path
-from typing import Callable
 
 import httpx
 import pytest
@@ -22,30 +22,42 @@ pytestmark = [
 
 
 def openai(http_client: httpx.AsyncClient, _tmp_path: Path) -> Model:
-    from pydantic_ai.models.openai import OpenAIModel
+    from pydantic_ai.models.openai import OpenAIChatModel
     from pydantic_ai.providers.openai import OpenAIProvider
 
-    return OpenAIModel('gpt-4o-mini', provider=OpenAIProvider(http_client=http_client))
+    return OpenAIChatModel('gpt-4o-mini', provider=OpenAIProvider(http_client=http_client))
 
 
-def gemini(http_client: httpx.AsyncClient, _tmp_path: Path) -> Model:
-    from pydantic_ai.models.gemini import GeminiModel
-    from pydantic_ai.providers.google_gla import GoogleGLAProvider
+def gemini(_: httpx.AsyncClient, _tmp_path: Path) -> Model:
+    from pydantic_ai.models.google import GoogleModel
 
-    return GeminiModel('gemini-1.5-pro', provider=GoogleGLAProvider(http_client=http_client))
+    return GoogleModel('gemini-1.5-pro')
 
 
 def vertexai(http_client: httpx.AsyncClient, tmp_path: Path) -> Model:
-    from pydantic_ai.models.gemini import GeminiModel
-    from pydantic_ai.providers.google_vertex import GoogleVertexProvider
+    from google.oauth2 import service_account
 
-    service_account_content = os.environ['GOOGLE_SERVICE_ACCOUNT_CONTENT']
-    service_account_path = tmp_path / 'service_account.json'
-    service_account_path.write_text(service_account_content)
-    return GeminiModel(
-        'gemini-1.5-flash',
-        provider=GoogleVertexProvider(service_account_file=service_account_path, http_client=http_client),
+    from pydantic_ai.models.google import GoogleModel
+    from pydantic_ai.providers.google import GoogleProvider
+
+    if service_account_path := os.environ.get('GOOGLE_APPLICATION_CREDENTIALS'):
+        project_id = json.loads(Path(service_account_path).read_text())['project_id']
+    elif service_account_content := os.environ.get('GOOGLE_SERVICE_ACCOUNT_CONTENT'):
+        project_id = json.loads(service_account_content)['project_id']
+        service_account_path = tmp_path / 'service_account.json'
+        service_account_path.write_text(service_account_content)
+    else:
+        pytest.skip(
+            'VertexAI live test requires GOOGLE_APPLICATION_CREDENTIALS or GOOGLE_SERVICE_ACCOUNT_CONTENT to be set'
+        )
+
+    credentials = service_account.Credentials.from_service_account_file(  # type: ignore[reportUnknownReturnType]
+        service_account_path,
+        scopes=['https://www.googleapis.com/auth/cloud-platform'],
     )
+    provider = GoogleProvider(credentials=credentials, project=project_id)
+    provider.client.aio._api_client._async_httpx_client = http_client  # type: ignore
+    return GoogleModel('gemini-2.0-flash', provider=provider)
 
 
 def groq(http_client: httpx.AsyncClient, _tmp_path: Path) -> Model:
@@ -63,10 +75,10 @@ def anthropic(http_client: httpx.AsyncClient, _tmp_path: Path) -> Model:
 
 
 def ollama(http_client: httpx.AsyncClient, _tmp_path: Path) -> Model:
-    from pydantic_ai.models.openai import OpenAIModel
+    from pydantic_ai.models.openai import OpenAIChatModel
     from pydantic_ai.providers.openai import OpenAIProvider
 
-    return OpenAIModel(
+    return OpenAIChatModel(
         'qwen2:0.5b', provider=OpenAIProvider(base_url='http://localhost:11434/v1/', http_client=http_client)
     )
 
@@ -87,13 +99,13 @@ def cohere(http_client: httpx.AsyncClient, _tmp_path: Path) -> Model:
 
 params = [
     pytest.param(openai, id='openai'),
-    pytest.param(gemini, marks=pytest.mark.skip(reason='API seems very flaky'), id='gemini'),
+    pytest.param(gemini, id='gemini', marks=pytest.mark.skip(reason='API seems very flaky')),
     pytest.param(vertexai, id='vertexai'),
-    pytest.param(groq, id='groq'),
+    pytest.param(groq, id='groq', marks=pytest.mark.skip(reason='test_structured has started failing')),
     pytest.param(anthropic, id='anthropic'),
     pytest.param(ollama, id='ollama'),
     pytest.param(mistral, id='mistral'),
-    pytest.param(cohere, id='cohere'),
+    pytest.param(cohere, id='cohere', marks=pytest.mark.skip(reason='Might be causing hangs in CI')),
 ]
 GetModel = Callable[[httpx.AsyncClient, Path], Model]
 
